#| -*-Scheme-*-

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015,
    2016, 2017, 2018, 2019 Matthew Birkholz

This file is part of a Pango plugin for MIT/GNU Scheme Pucked.

This plugin is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This plugin is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this plugin; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

|#

;;;; Pango Objects
;;; package: (gtk pango)

(C-include "pango")

(define-syntax error-if-null
  (syntax-rules ()
    ((_ alien message ...)
     (if (alien-null? alien)
	 (error message ...)))))

(define (color? object)
  (and (flo:flonum? object) (fix:= 4 (flo:vector-length object))))

(define-integrable (make-color) (flo:vector-cons 4))

(define-integrable (color-red o) (flo:vector-ref o 0))
(define-integrable (color-green o) (flo:vector-ref o 1))
(define-integrable (color-blue o) (flo:vector-ref o 2))
(define-integrable (color-alpha o) (flo:vector-ref o 3))

(define-integrable (set-color-red! o r) (flo:vector-set! o 0 r))
(define-integrable (set-color-green! o g) (flo:vector-set! o 1 g))
(define-integrable (set-color-blue! o b) (flo:vector-set! o 2 b))
(define-integrable (set-color-alpha! o a) (flo:vector-set! o 3 a))

(define (rgba-color red green blue alpha)
  (let ((color (make-color)))
    (set-color-red! color red)
    (set-color-green! color green)
    (set-color-blue! color blue)
    (set-color-alpha! color alpha)
    color))

(define white (rgba-color 1. 1. 1. 1.))
(define black (rgba-color 0. 0. 0. 1.))

(define (->color spec operator)
  (assert-glib-locked '->color)
  (cond ((color? spec) spec)
	((eq? spec 'white) white)
	((eq? spec 'black) black)
	((symbol? spec) (pango-color-parse (symbol->string spec)))
	((string? spec) (pango-color-parse spec))
	(else
	 (error:wrong-type-argument spec "a color spec" operator))))

(define (pango-color-parse spec)
  (guarantee string? spec 'pango-color-parse)
  (assert-glib-locked 'pango-color-parse)
  (let ((rgb (malloc (C-sizeof "PangoColor") '|PangoColor|)))
    (if (zero? (C-call "pango_color_parse" rgb (string->utf8 spec)))
	(error:wrong-type-argument spec "a color spec" 'pango-color-parse)
	(let ((color (make-color)))

	  (define-integrable (scale int)
	    (flo:/ (->flonum int)
		   (->flonum (shift-left 1 (* 8 (C-sizeof "guint16"))))))

	  (set-color-red! color (scale (C-> rgb "PangoColor red")))
	  (set-color-green! color (scale (C-> rgb "PangoColor green")))
	  (set-color-blue! color (scale (C-> rgb "PangoColor blue")))
	  (set-color-alpha! color 1.)
	  (free rgb)
	  color))))

(define-class (<pango-layout> (constructor ()))
    (<gobject>))

(define-guarantee pango-layout "a <pango-layout>")

(define-method initialize-instance ((pango-layout <pango-layout>))
  (call-next-method pango-layout)
  (set-alien/ctype! (gobject-alien pango-layout) '|PangoLayout|))

(define (pango-layout-get-context layout)
  (guarantee-pango-layout layout 'pango-layout-get-context)
  (assert-glib-locked 'pango-layout-get-context)
  (C-call "pango_layout_get_context"
	  (make-alien '|PangoContext|)
	  (gobject-alien layout)))

(define (pango-layout-context-changed layout)
  (guarantee-pango-layout layout 'pango-layout-context-changed)
  (assert-glib-locked 'pango-layout-context-changed)
  (C-call "pango_layout_context_changed" (gobject-alien layout)))

(define (pango-layout-get-font-description layout)
  (guarantee-pango-layout layout 'pango-layout-get-font-description)
  (assert-glib-locked 'pango-layout-get-font-description)
  (C-call "pango_layout_get_font_description"
	  (make-alien '|PangoFontDescription|)
	  (gobject-alien layout)))

(define (pango-layout-set-font-description layout font)
  ;; The toolkit makes a copy of font.
  (guarantee-pango-font-description font 'pango-layout-set-font-description)
  (assert-glib-locked 'pango-layout-set-font-description)
  (C-call "pango_layout_set_font_description"
	  (gobject-alien layout) (if (not font) 0 font)))

(define (pango-layout-set-text layout text)
  (guarantee-pango-layout layout 'pango-layout-set-text)
  (guarantee string? text 'pango-layout-set-text)
  (assert-glib-locked 'pango-layout-set-text)
  (let ((text-bv (string->utf8 text)))
    (C-call "pango_layout_set_text" (gobject-alien layout)
	    text-bv (bytevector-length text-bv))))

(define (pango-layout-set-markup layout markup)
  (guarantee-pango-layout layout 'pango-layout-set-markup)
  (guarantee string? markup 'pango-layout-set-markup)
  (assert-glib-locked 'pango-layout-set-markup)
  (let ((markup-bv (string->utf8 markup)))
    (C-call "pango_layout_set_markup" (gobject-alien layout)
	    markup-bv (bytevector-length markup-bv))))

(define (pango-layout-get-pixel-extents layout receiver)
  (guarantee-pango-layout layout 'pango-layout-set-text)
  (guarantee-procedure-of-arity receiver 2 'pango-layout-set-text)
  (assert-glib-locked 'pango-layout-get-pixel-extents)
  (let (;(ink-extent (pango-rectangle))
	(logical-extent (pango-rectangle)))
    (C-call "pango_layout_get_pixel_extents"
	    (gobject-alien layout)
	    0 ;ink-extent
	    logical-extent)
    ;; Can ink extend beyond the logical extent?
    (let ((width (C-> logical-extent "PangoRectangle width"))
	  (height (C-> logical-extent "PangoRectangle height")))
      ;(free ink-extent)
      (free logical-extent)
      (receiver width height))))

(define (pango-layout-index-to-pos layout index receiver)
  (guarantee-pango-layout layout 'pango-layout-index-to-pos)
  (guarantee-procedure-of-arity receiver 4 'pango-layout-index-to-pos)
  (assert-glib-locked 'pango-layout-index-to-pos)
  (let ((rect (pango-rectangle)))
    (C-call "pango_layout_index_to_pos" (gobject-alien layout) index rect)
    (let ((x (pangos->pixels (C-> rect "PangoRectangle x")))
	  (y (pangos->pixels (C-> rect "PangoRectangle y")))
	  (width (pangos->pixels (C-> rect "PangoRectangle width")))
	  (height (pangos->pixels (C-> rect "PangoRectangle height"))))
      (free rect)
      (receiver x y width height))))

(define (pango-layout-xy-to-index layout x y)
  (guarantee-pango-layout layout 'pango-layout-xy-to-index)
  (assert-glib-locked 'pango-layout-xy-to-index)
  (let ((index-alien (malloc (C-sizeof "int") 'int)))
    (if (fix:= 0 (C-call "pango_layout_xy_to_index"
			 (gobject-alien layout)
			 (pixels->pangos x) (pixels->pangos y)
			 index-alien 0))
	(begin
	  (free index-alien)
	  #f)
	(let ((index (C-> index-alien "int")))
	  (free index-alien)
	  index))))

(define (pango-layout-get-baseline layout)
  (guarantee-pango-layout layout 'pango-layout-get-baseline)
  (assert-glib-locked 'pango-layout-get-baseline)
  (let ((iter (make-alien '|PangoLayoutIter|))
	(copy (make-alien '|PangoLayoutIter|)))
    (add-glib-cleanup iter (make-pango-layout-iter-cleanup copy))
    (C-call "pango_layout_get_iter" copy (gobject-alien layout))
    (copy-alien-address! iter copy)
    (error-if-null iter "Could not create:" iter layout)
    (let ((baseline
	   (pangos->pixels
	    (C-call "pango_layout_iter_get_baseline" iter))))
      (pango-layout-iter-free iter)
      baseline)))

(define (make-pango-layout-iter-cleanup alien)
  (named-lambda (pango-layout-iter-cleanup)
    (assert-glib-locked 'pango-layout-iter-cleanup)
    (assert-without-interruption 'pango-layout-iter-cleanup)
    (if (not (alien-null? alien))
	(begin
	  (C-call "pango_layout_iter_free" alien)
	  (alien-null! alien)))))

(define (pango-layout-iter-free iter)
  (assert-glib-locked 'pango-layout-iter-free)
  (without-interruption
   (lambda ()
     (if (not (alien-null? iter))
	 (begin
	   (execute-glib-cleanup iter)
	   (alien-null! iter))))))

;;; PangoFontDescription

(define (pango-font-description-from-string string)
  ;; The returned PangoFontDescription is owned by Scheme.
  (guarantee string? string 'pango-font-description-from-string)
  (assert-glib-locked 'pango-font-description-from-string)
  (let ((font (make-alien '|PangoFontDescription|))
	(copy (make-alien '|PangoFontDescription|))
	(string-bv (string->utf8 string)))
    (add-glib-cleanup font (make-pango-font-description-cleanup copy))
    (C-call "pango_font_description_from_string" copy string-bv)
    (copy-alien-address! font copy)
    (error-if-null font "Could not create:" font)
    font))

(define (make-pango-font-description-cleanup alien)
  (named-lambda (pango-font-description-cleanup)
    (assert-glib-locked 'pango-font-description-cleanup)
    (assert-without-interruption 'pango-font-description-cleanup)
    (if (not (alien-null? alien))
	(begin
	  (C-call "pango_font_description_free" alien)
	  (alien-null! alien)))))

(define (pango-font-description-free font)
  (guarantee-pango-font-description font 'pango-font-description-free)
  (assert-glib-locked 'pango-font-description-free)
  (without-interruption
   (lambda ()
     (if (not (alien-null? font))
	 (begin
	   (execute-glib-cleanup font)
	   (alien-null! font))))))

(define (pango-font-description-to-string font)
  (guarantee-pango-font-description font 'pango-font-description-to-string)
  (assert-glib-locked 'pango-font-description-to-string)
  (if (alien-null? font)
      "<null>"
      (without-interruption
       (lambda ()
	 (let ((cstr (make-alien '|char|)))
	   (C-call "pango_font_description_to_string" cstr font)
	   (let ((str (c-peek-cstring cstr)))
	     (C-call "g_free" cstr)
	     str))))))

(define (pango-font-description-copy font)
  (guarantee-pango-font-description font 'pango-font-description-copy)
  (assert-glib-locked 'pango-font-description-copy)
  (let ((new (make-alien '|PangoFontDescription|))
	(copy (make-alien '|PangoFontDescription|)))
    (add-glib-cleanup new (make-pango-font-description-cleanup copy))
    (C-call "pango_font_description_copy" copy font)
    (copy-alien-address! new copy)
    (error-if-null new "Could not create:" new font)
    new))

(define-integrable (guarantee-pango-font-description object operator)
  (if (not (and (alien? object)
		(eq? '|PangoFontDescription| (alien/ctype object))))
      (error:wrong-type-argument
       object "a PangoFontDescription address" operator)))

;;; PangoContext

(define (pango-context-get-font-description context)
  ;; The returned PangoFontDescription is owned by the toolkit
  ;; (the PangoContext), not Scheme.
  (guarantee-pango-context context 'pango-context-get-font-description)
  (assert-glib-locked 'pango-context-get-font-description)
  (C-call "pango_context_get_font_description"
	  (make-alien '|PangoFontDescription|) context))

(define (pango-context-set-font-description context font)
  ;; Font is still owned by Scheme.  The toolkit makes a copy.
  (guarantee-pango-context context 'pango-context-set-font-description)
  (guarantee-pango-font-description font 'pango-context-set-font-description)
  (assert-glib-locked 'pango-context-set-font-description)
  (C-call "pango_context_set_font_description" context font))

(define (pango-context-get-metrics context font)
  ;; The new PangoFontMetrics is owned by Scheme.
  (guarantee-pango-context context 'pango-context-get-metrics)
  (guarantee-pango-font-description font 'pango-context-get-metrics)
  (assert-glib-locked 'pango-context-get-metrics)
  (let ((alien (make-alien '|PangoFontMetrics|))
	(copy (make-alien '|PangoFontMetrics|)))
    (add-glib-cleanup alien (make-pango-font-metrics-cleanup copy))
    (C-call "pango_context_get_metrics" copy context font 0)
    (copy-alien-address! alien copy)
    (error-if-null alien "Could not get:" alien)
    alien))

(define (pango-context-spacing context)
  (guarantee-pango-context context 'pango-context-spacing)
  (assert-glib-locked 'pango-context-spacing)
  (let ((layout (make-alien '|PangoLayout|)))
    (C-call "pango_layout_new" layout context)
    (let ((spacing (C-call "pango_layout_get_spacing" layout)))
      (C-call "g_object_unref" layout)
      spacing)))

(define-integrable (guarantee-pango-context object operator)
  (if (not (and (alien? object) (eq? '|PangoContext| (alien/ctype object))))
      (error:wrong-type-argument
	object "a PangoContext address" operator)))

;;; PangoFontMetrics

(define (make-pango-font-metrics-cleanup alien)
  (named-lambda (pango-font-metrics-cleanup)
    (assert-glib-locked 'pango-font-metrics-cleanup)
    (assert-without-interruption 'pango-font-metrics-cleanup)
    (if (not (alien-null? alien))
	(begin
	  (C-call "pango_font_metrics_unref" alien)
	  (alien-null! alien)))))

(define (pango-font-metrics-unref metrics)
  (guarantee-pango-font-metrics metrics 'pango-font-metrics-unref)
  (assert-glib-locked 'pango-font-metrics-unref)
  (without-interruption
   (lambda ()
     (if (not (alien-null? metrics))
	 (begin
	   (execute-glib-cleanup metrics)
	   (alien-null! metrics))))))

(define (pango-font-metrics-get-ascent metrics)
  (guarantee-live-pango-font-metrics metrics 'pango-font-metrics-get-ascent)
  (assert-glib-locked 'pango-font-metrics-get-ascent)
  (C-call "pango_font_metrics_get_ascent" metrics))

(define (pango-font-metrics-get-descent metrics)
  (guarantee-live-pango-font-metrics metrics 'pango-font-metrics-get-descent)
  (assert-glib-locked 'pango-font-metrics-get-descent)
  (C-call "pango_font_metrics_get_descent" metrics))

(define (pango-font-metrics-get-approximate-char-width metrics)
  (guarantee-live-pango-font-metrics metrics 'pango-font-metrics-get-approximate-char-width)
  (assert-glib-locked 'pango-font-metrics-get-approximate-char-width)
  (C-call "pango_font_metrics_get_approximate_char_width" metrics))

(define-integrable (guarantee-live-pango-font-metrics object operator)
  (guarantee-pango-font-metrics object operator)
  (if (alien-null? object) (error "Dead:" object)))

(define-integrable (guarantee-pango-font-metrics object operator)
  (if (not (and (alien? object) (eq? '|PangoFontMetrics| (alien/ctype object))))
      (error:wrong-type-argument
       object "a PangoFontMetrics address" operator)))

;;; PangoRectangle

(define (pango-rectangle #!optional x y width height)
  (if (default-object? x)
      (malloc (C-sizeof "PangoRectangle") '|PangoRectangle|)
      (let ((rect (malloc (C-sizeof "PangoRectangle") '|PangoRectangle|)))
	(C->= rect "PangoRectangle x" x)
	(C->= rect "PangoRectangle y" y)
	(C->= rect "PangoRectangle width" width)
	(C->= rect "PangoRectangle height" height)
	rect)))

(define-integrable (pangos->pixels pango-units)
  (quotient (int:+ pango-units 512) 1024))

(define-integrable (pixels->pangos pixel-units)
  (* pixel-units 1024))

;;; Debugging hacks.  No cleanups!  No locking!

(define (pango-context-list-families PangoContext)
  (let ((data-arg (malloc (C-sizeof "*") '(* (* |PangoFontFamily|))))
	(count-arg (malloc (C-sizeof "int") 'int)))
    (C-call "pango_context_list_families" PangoContext data-arg count-arg)
    (let ((data (C-> data-arg "*" (make-alien '(* |PangoFontFamily|))))
	  (count (C-> count-arg "int")))
      (free data-arg) (free count-arg)
      (let* ((scan (copy-alien data))
	     (family (make-alien '|PangoFontFamily|))
	     (alist
	      (let loop ((i 0) (entries '()))
		(if (fix:< i count)
		    (begin
		      (C-> scan "*" family)
		      (alien-byte-increment! scan (C-sizeof "*"))
		      (loop (fix:1+ i)
			    (cons
			     (cons* (pango-font-family-get-name family)
				    (pango-font-family-is-monospace? family)
				    (pango-font-family-faces family))
			     entries)))
		    entries))))
	(C-call "g_free" data)
	alist))))

(define (pango-font-family-get-name PangoFontFamily)
  (let ((name (make-alien '(const char))))
    (C-call "pango_font_family_get_name" name PangoFontFamily)
    (c-peek-cstring name)))

(define (pango-font-family-is-monospace? PangoFontFamily)
  (not (fix:zero? (C-call "pango_font_family_is_monospace" PangoFontFamily))))

(define (pango-font-family-faces PangoFontFamily)
  (let ((data-arg (malloc (C-sizeof "*") '(* (* |PangoFontFace|))))
	(count-arg (malloc (C-sizeof "int") 'int)))
    (C-call "pango_font_family_list_faces" PangoFontFamily data-arg count-arg)
    (let ((data (C-> data-arg "*" (make-alien '(* |PangoFontFace|))))
	  (count (C-> count-arg "int")))
      (free data-arg) (free count-arg)
      (let* ((scan (copy-alien data))
	     (face (make-alien '|PangoFontFace|))
	     (faces
	      (let loop ((i 0) (faces '()))
		(if (fix:< i count)
		    (begin
		      (C-> scan "*" face)
		      (alien-byte-increment! scan (C-sizeof "*"))
		      (loop (fix:1+ i)
			    (cons
			     (pango-font-face-get-name face)
			     faces)))
		    faces))))
	(C-call "g_free" data)
	faces))))

(define (pango-font-face-get-name PangoFontFace)
  (let ((name (make-alien '(const char))))
    (C-call "pango_font_face_get_face_name" name PangoFontFace)
    (c-peek-cstring name)))